/*
 * A mini C-like language scanner.
 */

#include <stdio.h>

#define IDENT_BUFLEN 256

struct clang
{
	char identBuf[IDENT_BUFLEN+1];
	int identLen;
	int curLine;

	int curs;
};

%% clang
{
	# Initialization code that will go into the ClangInit routine.
	init {
		fsm->identLen = 0;
		fsm->curLine = 1;
	}

	# Function to buffer a character.
	action bufChar {
		if ( fsm->identLen < IDENT_BUFLEN ) {
			fsm->identBuf[fsm->identLen] = fc;
			fsm->identLen += 1;
		}
	}

	# Function to clear the buffer.
	action clearBuf {
		fsm->identLen = 0;
	}

	# Functions to dump tokens as they are matched.
	action ident {
		fsm->identBuf[fsm->identLen] = 0;
		printf("ident(%i): %s\n", fsm->curLine, fsm->identBuf);
	}
	action literal {
		fsm->identBuf[fsm->identLen] = 0;
		printf("literal(%i): \"%s\"\n", fsm->curLine, fsm->identBuf);
	}
	action float {
		fsm->identBuf[fsm->identLen] = 0;
		printf("float(%i): %s\n", fsm->curLine, fsm->identBuf);
	}
	action int {
		fsm->identBuf[fsm->identLen] = 0;
		printf("int(%i): %s\n", fsm->curLine, fsm->identBuf);
	}
	action hex {
		fsm->identBuf[fsm->identLen] = 0;
		printf("hex(%i): 0x%s\n", fsm->curLine, fsm->identBuf);
	}
	action symbol {
		fsm->identBuf[fsm->identLen] = 0;
		printf("symbol(%i): %s\n", fsm->curLine, fsm->identBuf);
	}

	# Alpha numberic characters or underscore.
	alnumu = alnum | '_';

	# Alpha charactres or underscore.
	alphau = alpha | '_';

	# Symbols. Upon entering clear the buffer. On all transitions
	# buffer a character. Upon leaving dump the symbol.
	symbol = ( punct - [_'"] ) >clearBuf $bufChar %symbol;

	# Identifier. Upon entering clear the buffer. On all transitions
	# buffer a character. Upon leaving, dump the identifier.
	ident = (alphau . alnumu*) >clearBuf $bufChar %ident;

	# Match single characters inside literal strings. Or match 
	# an escape sequence. Buffers the charater matched.
	sliteralChar = 
			( extend - ['\\] ) @bufChar |
			( '\\' . extend @bufChar );
	dliteralChar = 
			( extend - ["\\] ) @bufChar |
			( '\\' . extend @bufChar );

	# Single quote and double quota literals. At the start clear
	# the buffer. Upon leaving dump the literal.
	sliteral = ('\'' @clearBuf . sliteralChar* . '\'' ) %literal;
	dliteral = ('"' @clearBuf . dliteralChar* . '"' ) %literal;
	literal = sliteral | dliteral;

	# Whitespace is standard ws, newlines and control codes.
	whitespace = any - 0x21..0x7e;

	# Describe both c style comments and c++ style comments. The
	# priority bump on tne terminator of the comments brings us
	# out of the extend* which matches everything.
	ccComment = '//' . extend* $0 . '\n' @1;
	cComment = '/*' . extend* $0 . '*/' @1;

	# Match an integer. We don't bother clearing the buf or filling it.
	# The float machine overlaps with int and it will do it.
	int = digit+ %int;

	# Match a float. Upon entering the machine clear the buf, buffer
	# characters on every trans and dump the float upon leaving.
	float =  ( digit+ . '.' . digit+ ) >clearBuf $bufChar %float;

	# Match a hex. Upon entering the hex part, clear the buf, buffer characters
	# on every trans and dump the hex on leaving transitions.
	hex = '0x' . xdigit+ >clearBuf $bufChar %hex;

	# Or together then star all the lanuage elements. Use the longest match
	# kleene star operator This is so that when we see 'aa' we stay in the fin
	# machine to match an ident of length two and not wrap around to the front
	# to match two idents of length one.
	clang_main = ( 
		ccComment |
		cComment |
		symbol |
		ident |
		literal |
		whitespace |
		int |
		float |
		hex )**;

	# This machine matches everything, taking note of newlines.
	newline = ( any | '\n' @{ fsm->curLine += 1; } )*;

	# The final fsm is the lexer intersected with the newline machine which
	# will count lines for us. Since the newline machine accepts everything,
	# the strings accepted is goverened by the clang_main machine, onto which
	# the newline machine overlays line counting.
	main := clang_main & newline;
}

#define BUFSIZE 2048

int main()
{
	char buf[BUFSIZE];
	struct clang clang;

	clang_init( &clang );
	while ( 1 ) {
		int len = fread( buf, 1, BUFSIZE, stdin );

		clang_execute( &clang, buf, len );

		if ( len != BUFSIZE )
			break;
	}
	clang_finish( &clang );
	return 0;
}

